"""LoginLockout
"""

__author__ = "Dylan Jay <software@pretaweb.com>"

from AccessControl.Permissions import add_user_folders
from Products.PluggableAuthService import registerMultiPlugin
from plugin import LoginLockout, \
                   manage_addLoginLockout, \
                   manage_addLoginLockoutForm

from Products.CMFCore.DirectoryView import registerDirectory

from Products.CMFCore import utils
from Products.LoginLockout.plugin import PROJECTNAME
from Products.LoginLockout.loginlockout_tool import LoginLockoutTool

def initialize(context):
    """Initialize the LoginLockout plugin.
    Register skin directory.
    """
    registerMultiPlugin(LoginLockout.meta_type)
    
    context.registerClass(LoginLockout,
                          permission=add_user_folders,
                          constructors=(manage_addLoginLockoutForm,
                                        manage_addLoginLockout),
                          #icon='www/noduplicatelogin.png',
                          visibility=None,
                          )

    # register the custom skins directory
    GLOBALS = globals()
    registerDirectory('skins', GLOBALS)

    # register the tool
    tools = ( LoginLockoutTool, )
    utils.ToolInit(
        PROJECTNAME,
        tools=tools).initialize(context)




from plugin import PROJECTNAME

TOOL_ID = "loginlockout_tool"

CONFIGLETS = (
    {
        'id'         : 'LoginLockoutConf',
        'name'       : 'LoginLockout',
        'action'     : 'string:${portal_url}/loginlockout_settings',
        'condition'  : '',
        'category'   : 'Products',
        # section to which the configlet should be added:
        # (Plone,Products,Members)
        'visible'    : 1,
        'appId'      : PROJECTNAME,
        'permission' : 'ManagePortal',
        'imageUrl'   : 'lock_icon.gif',
    },   
) 




"""LoginLockout.
   Locks out the user when they make too many different unsuccessful login attempts.
   
   An AuthenticateUser plugin increments a count for each login with a different
   password.
   
   A UpdateCredentials plugin resets that count as this indicates a successful login.
   
   If the count reaches the max then AuthenticateUser Plugin throws a Unauthorised 
   exception.
   
   The challenge machinery is inacted and a Challenge plugin recognises the user is locked
   out and redirects them to a page informing them they are locked out and to contact the 
   admin
   
   The admin can view and reset attempts via the ZMI at any time
"""

__author__ = "Dylan Jay <software@pretaweb.com>"

import logging

from urllib import quote, unquote

from BTrees.OOBTree import OOBTree
from DateTime import DateTime
from AccessControl import ClassSecurityInfo, Permissions
from AccessControl.Permissions import view
from AccessControl import AuthEncoding
from Globals import InitializeClass
from OFS.Cache import Cacheable
from OFS.Folder import Folder

from zExceptions import Unauthorized

from Products.PluggableAuthService import PluggableAuthService
from Products.PluggableAuthService.plugins.BasePlugin import BasePlugin
from Products.PluggableAuthService.permissions import ManageUsers
from Products.PluggableAuthService.utils import classImplements
from Products.PluggableAuthService.interfaces.plugins import \
    IAuthenticationPlugin, \
    ICredentialsUpdatePlugin, \
    ICredentialsResetPlugin, \
    IChallengePlugin, \
    IAnonymousUserFactoryPlugin


from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Products.PageTemplates.ZopePageTemplate import ZopePageTemplate

log = logging.getLogger('LoginLockout')

manage_addLoginLockoutForm = PageTemplateFile(
    'www/loginLockoutAdd',
    globals(),
    __name__='manage_addLoginLockoutForm' )

def manage_addLoginLockout(dispatcher,
                               id,
                               title=None,
                               REQUEST=None):
    """Add a LoginLockout plugin to a Pluggable Auth Service."""

    obj = LoginLockout(id, title
                           )
    dispatcher._setObject(obj.getId(), obj)

    if REQUEST is not None:
        REQUEST['RESPONSE'].redirect('%s/manage_workspace?manage_tabs_message='
                                     'LoginLockout+plugin+added.'
                                     % dispatcher.absolute_url())

class LoginLockout(Folder, BasePlugin, Cacheable):

    """PAS plugin that rejects logins after X attemps
    """

    lockout_path = 'lockout'
    meta_type = 'Login Lockout Plugin'
    cookie_name = '__noduplicate'
    security = ClassSecurityInfo()

    _properties = ( { 'id'    : 'title'
                    , 'label' : 'Title'
                    , 'type'  : 'string'
                    , 'mode'  : 'w'
                    }
                  , { 'id'    : '_max_attempts'
                    , 'label' : 'Number of Allowed Attempts'
                    , 'type'  : 'int'
                    , 'mode'  : 'w'
                    }
                  , { 'id'    : '_reset_period'
                    , 'label' : 'Attempt Reset Period (hours)'
                    , 'type'  : 'float'
                    , 'mode'  : 'w'
                    }
                  )

    def __init__(self, id, title=None):
        self._id = self.id = id
        self.title = title
        
        self._login_attempts = OOBTree() # userid : (Count:int, DateTime, IP:string)
        self._reset_period = 24.0
        self._max_attempts = 3

    security.declarePrivate('authenticateCredentials')
    def authenticateCredentials(self, credentials):
        """See IAuthenticationPlugin.

        This plugin will actually never authenticate.

        o We expect the credentials to be those returned by
          ILoginPasswordExtractionPlugin.
        """
        request = self.REQUEST
        response = request['RESPONSE']
        pas_instance = self._getPAS()

        login = credentials.get('login')
        password = credentials.get('password')

        if None in (login, password, pas_instance):
            return None

        if self.isLockedout(login):
            #self.resetAllCredentials(request, response)
#            credentials['login'] = '' # pretty dodgy but not sure how else to do it
            request['portal_status_message'] = (
                    "This account is locked."
                    "Please contact your administrator to unlock this account")
            request['locked_login'] = login # so challenge plugin can fire
            request.set('__ac','') #HACK - need ot reset in current request not just reponse like cookie auth does
            self.resetAllCredentials(request, response) # must reset so we don't lockout of the login page
            count,last,IP = self.getAttempts(login)
            log.info("Attempt denied due to lockout: %s, %s ",login, IP)
            raise Unauthorized

        #record login so challange plugin can use it
        #self._t_attempted_logins = getattr(self,'_t_attempted_logins',[]) + [login]
        
#        self.setAttempt(login, password) # set as failed attempt and reset if we get updateCredentuals
        request.set('attempted_logins',(login,password))

        return None # Note that we never return anything useful


    security.declarePrivate('createAnonymousUser')
    def createAnonymousUser(self):
        """ if we got anon then attempt failed """
        login,password = self.REQUEST.get('attempted_logins', ('',''))
        if login:
            self.setAttempt(login, password)
            log.info("Failed login attempt: %s ",login)
        


    security.declarePrivate('updateCredentials')
    def updateCredentials(self, request, response, login, new_password ):

        """ Called on succesful attempt. reset user
        """
        self.resetAttempts(login, new_password)
        log.info("Successful login: %s ",login)

    security.declarePrivate('challenge')
    def challenge(self, request, response, **kw):
        """ Challenge the user for credentials. """
        login = request.get('locked_login',None)
        if login and self.isLockedout(login):
            return self.unauthorized()
        return 0

    security.declarePrivate('unauthorized')
    def unauthorized(self):
        req = self.REQUEST
        resp = req['RESPONSE']

        # Redirect if desired.
        url = self.getLockoutURL()
        if url is not None:
            resp.redirect(url, lock=1)
            return 1

        # Could not challenge.
        return 0

    security.declarePrivate('getLockoutURL')
    def getLockoutURL(self):
        """ Where to send people for logging in """
        if self.lockout_path.startswith('/'):
            return self.lockout_path
        elif self.lockout_path != '':
            return '%s/%s' % (self.absolute_url(), self.lockout_path)
        else:
            return None

    security.declarePrivate('getRootPlugin')
    def getRootPlugin(self):
        #pas = self.superValues('PluggableAuthService')
        #if pas:
            
            #pas = pas[-1]
            pas = self.getPhysicalRoot().acl_users
            plugins = pas.objectValues([self.meta_type])
            if plugins:
                return plugins[0]
            


    security.declarePrivate('setAttempt')
    def setAttempt(self, login, password):
        "increment attempt count and record date stamp last attempt and IP"
        root = self.getRootPlugin()
        count,last,IP,reference = root._login_attempts.get(login, (0, None, '', None))
        
        if reference and AuthEncoding.pw_validate( reference, password ):
            return # we don't count repeating same password in case its correct
#        elif last and (DateTime() - last)/24.0 > self._reset_period:
#            count = 1
        else:
            count += 1
        IP = self.REQUEST.get('HTTP_X_FORWARDED_FOR','')
        if not IP:
            IP = self.REQUEST.get('REMOTE_ADDR','')
        log.info("user '%s' attempt# %i %s last: %s", login, count, IP, last)
        last = DateTime()
        reference = AuthEncoding.pw_encrypt( password )
        root._login_attempts[login] = (count,last,IP, reference)
        
    security.declarePrivate('getAttempts')
    def getAttempts(self, login):
        "return the count, last attempt datestamp and IP of last attempt"
        root = self.getRootPlugin()
        count,last,IP, pw_hash = root._login_attempts.get(login, (0, None, '',''))
        return count,last,IP

    security.declarePrivate('isLockedout')
    def isLockedout(self, login):
        root = self.getRootPlugin()
        count,last,IP = root.getAttempts(login)
        return count >= root._max_attempts        


    security.declarePrivate('resetAttempts')
    def resetAttempts(self, login, password=None ):
        "reset to zero and update pw referece so same attempts pass"
        root = self.getRootPlugin()
        if root._login_attempts.get(login, None):
            del root._login_attempts[login]

    security.declarePrivate('resetAllCredentials')
    def resetAllCredentials(self, request, response):
        """Call resetCredentials of all plugins.

        o This is not part of any contract.
        """
        # This is arguably a bit hacky, but calling
        # pas_instance.resetCredentials() will not do anything because
        # the user is still anonymous.  (I think it should do
        # something nevertheless.)
        pas_instance = self._getPAS()
        plugins = pas_instance._getOb('plugins')
        cred_resetters = plugins.listPlugins(ICredentialsResetPlugin)
        for resetter_id, resetter in cred_resetters:
            resetter.resetCredentials(request, response)


    #
    #   ZMI
    #
    manage_options = ( ( { 'label': 'Users', 
                           'action': 'manage_users', }
                         ,
                       )
                     + BasePlugin.manage_options[:1]
                     + Folder.manage_options[:1]
                     + Folder.manage_options[2:]
                     )


    security.declareProtected( ManageUsers, 'manage_users' )
    manage_users = PageTemplateFile( 'www/llLockouts'
                                   , globals()
                                   , __name__='manage_users'
                                   )

    security.declarePrivate('manage_afterAdd')
    def manage_afterAdd(self, item, container):
        """ Setup tasks upon instantiation """
        if not 'lockout' in self.objectIds():
            lockout = ZopePageTemplate( id='lockout'
                                           , text=BASIC_LOCKOUT
                                           )
            lockout.title = 'User Lockout'
            lockout.manage_permission(view, roles=['Anonymous'], acquire=1)
            self._setObject( 'lockout', lockout, set_owner=0 )


    security.declareProtected( ManageUsers, 'manage_resetUsers' )
    def manage_resetUsers( self
                      , logins
                      , RESPONSE=None
                      ):
        """ Reset lockout so user can login again
        """
        for login in logins:
            self.resetAttempts(login)
        message = "User reset"
        if RESPONSE is not None:
            RESPONSE.redirect( '%s/manage_users?manage_tabs_message=%s'
                             % ( self.absolute_url(), message )
                             )

    security.declareProtected( ManageUsers, 'getAttemptInfo' )
    def getAttemptInfo( self, login ):

        """ user_id -> {}
        """
        count,last,IP = self.getAttempts(login)
        return { 'login' : login
               , 'last' : last
               , 'IP' : IP
               , 'count' : count
               }

    security.declareProtected( ManageUsers, 'listAttempts' )
    def listAttempts( self ):

        """ -> ( {}, ...{} )

        o Return one mapping per user, with the following keys
        """
        root = self.getRootPlugin()
        return [ self.getAttemptInfo( x ) for x in root._login_attempts.keys() ]



classImplements(LoginLockout,
                ICredentialsUpdatePlugin,
                IAuthenticationPlugin,
                IChallengePlugin,
                IAnonymousUserFactoryPlugin)

InitializeClass(LoginLockout)

PROJECTNAME = 'LoginLockout'
PLUGIN_ID = 'login_lockout_plugin'
PLUGIN_TITLE = 'Disable account after failed login attempts.'
BASIC_LOCKOUT = """<html>
  <head>
    <title> User Lockedout </title>
  </head>

  <body>

    <h3> Account has been locked </h3>

<p>
  This account has now been locked for security purposes. Please
  contact your administrator to unlock this account
</p>


  </body>

</html>
"""




import unittest
import os, sys, tempfile, shutil, StringIO

from zope.testing import doctest
from zope.testing import doctestunit
from zope.component import testing
#import DateTime
#from DateTime.interfaces import DateTimeError, SyntaxError, DateError, TimeError
from Testing import ZopeTestCase as ztc
from Products.Five import zcml
from Products.Five import fiveconfigure
from Products.PloneTestCase import PloneTestCase as ptc
from Products.PloneTestCase.layer import onsetup
from Products.Five.testbrowser import Browser
from zope.component import getUtility
from Products.CMFCore.utils import getToolByName
import DateTime

print "Start installProduct"
ztc.installProduct('LoginLockout')
print "Finished installProduct setup (hopefully)"

class TestCase(ptc.PloneTestCase):
    """ We use this base class for all the tests in this package. If necessary,
        we can put common utility or setup code in here. This applies to unit
        test cases. """
    _configure_portal = False

    def beforeTearDown(self):
        pass

    def afterSetUp(self):

        self.portal.acl_users.portal_role_manager.updateRolesList()

        self.portal.acl_users._doAddUser('admin', 'admin', ('Manager',), [])
        self.portal.acl_users._doAddUser('user', 'user', ('Member',), [])

        self.portal.error_log._ignored_exceptions = ()


@onsetup
def setup_product():
    """ """
    fiveconfigure.debug_mode = True
    import Products.LoginLockout
    zcml.load_config('configure.zcml', Products.LoginLockout)
    fiveconfigure.debug_mode = False

    #print "Start setupProduct"
    ztc.installPackage('Products.LoginLockout')
    #print "Finished setupProduct (hopefully)"


setup_product()
ptc.setupPloneSite(extension_profiles=(), with_default_memberarea=False,
        products=['Products.LoginLockout'])


def test_suite():
    return unittest.TestSuite([
        # USE CASES
        ztc.FunctionalDocFileSuite(
            usecase , package='Products.LoginLockout', test_class=TestCase,
            optionflags=doctest.REPORT_ONLY_FIRST_FAILURE |
                        doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS)
            for usecase in ['README.txt',
                           ]

        ])

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



from Globals import InitializeClass
from OFS.SimpleItem import SimpleItem
from AccessControl import ClassSecurityInfo
from AccessControl.SecurityManagement import getSecurityManager
#from permissions import ManagePortal

from Products.CMFCore.utils import UniqueObject, getToolByName

from Products.LoginLockout.config import TOOL_ID
from Products.LoginLockout.plugin import PROJECTNAME

class LoginLockoutTool(UniqueObject,  SimpleItem):
    """ A tool to facilitate calling the plugin's methods for viewing login
    attempts, resetting attempts, etc. from a view template.
    """
    meta_type = PROJECTNAME
    id = TOOL_ID
    title = "Login Lockout Tool"
    security = ClassSecurityInfo()

    security.declarePrivate('_getPlugin')
    def _getPlugin(self):
        acl_users = getToolByName(self, 'acl_users')
        return acl_users.login_lockout_plugin

    #security.declareProtected(ManagePortal, 'listAttempts')
    def listAttempts(self):
        return self._getPlugin().listAttempts()

    def manage_resetUsers(self, logins, RESPONSE=None):
        return self._getPlugin().manage_resetUsers(logins, RESPONSE=None)

InitializeClass(LoginLockoutTool)



from StringIO import StringIO
from Products.CMFCore.utils import getToolByName
from Products.PlonePAS.Extensions.Install import activatePluginInterfaces
from Products.LoginLockout.plugin import PROJECTNAME, PLUGIN_ID, PLUGIN_TITLE
from Products.CMFCore.permissions import ManagePortal

# for installing skin
from Products.CMFCore.DirectoryView import addDirectoryViews

# for adding tool
from Products.LoginLockout.config import TOOL_ID, CONFIGLETS

def setupLoginLockout(context):
    if context.readDataFile('loginlockout.txt') is None:
        return
    site = context.getSite()
    install(site)

def removeLoginLockout(context):
    if context.readDataFile('loginlockout.txt') is None:
        return
    site = context.getSite()
    uninstall(site)

def install( portal ):

    """ This plugin needs to be installed in two places, the instance PAS where
    logins occur and the root acl_users.

    Different interfaces need to be activated for either case.
    """
    out = StringIO()
    print >> out, "Installing %s:" % PROJECTNAME

    plone_pas = getToolByName(portal, 'acl_users')
    zope_pas = portal.getPhysicalRoot().acl_users

    # define the interfaces which need to be activated for either PAS
    interfaces_for_paservices = {
        plone_pas: ['IAuthenticationPlugin','IChallengePlugin',
                'ICredentialsUpdatePlugin'],
        zope_pas: ['IChallengePlugin','IAnonymousUserFactoryPlugin'],
        }
    for (pas, interfaces) in interfaces_for_paservices.iteritems():
        registry = pas.plugins
        existing = pas.objectIds()
        if PLUGIN_ID not in existing:
            loginlockout = pas.manage_addProduct[PROJECTNAME]
            loginlockout.manage_addLoginLockout(PLUGIN_ID, PLUGIN_TITLE)
            activatePluginSelectedInterfaces(pas, PLUGIN_ID, out, interfaces)

    # define which interfaces need to be moved to top of plugin list
    move_to_top_for = {
        plone_pas: 'IChallengePlugin',
        zope_pas: 'IAnonymousUserFactoryPlugin',
        }
    for (pas, interface) in move_to_top_for.iteritems():
        movePluginToTop(pas, PLUGIN_ID, interface, out)
    
    # install skins layer
    installSkin(portal, out, globals(), 'loginlockout_templates')

    # add tool
    addTool(portal, PROJECTNAME, TOOL_ID)

    # install configlet
    installConfiglets(portal, out, CONFIGLETS)

    print >> out, "Successfully installed %s." % PROJECTNAME
    return out.getvalue()

def uninstall( portal ):
    out = StringIO()
    print >> out, "Uninstalling %s:" % PROJECTNAME
    plone_pas = getToolByName(portal, 'acl_users')
    zope_pas = portal.getPhysicalRoot().acl_users
    for pas in [plone_pas, zope_pas]:
        existing = pas.objectIds()
        if PLUGIN_ID in existing:
            pas.manage_delObjects(PLUGIN_ID)

    # uninstall configlets
    installConfiglets(portal, out, CONFIGLETS, uninstall=True)

def activatePluginSelectedInterfaces(pas, plugin, out, selected_interfaces, 
        disable=[]):
    """This is derived from PlonePAS's activatePluginInterfaces, but 
    will only activate the plugin for selected interfaces. 
    
    For LoginLockout, you'll want to activate different interfaces for 
    different PAS instances.

    The PAS instance (either Plone's or Zope's) in which to activate the 
    interfaces is passed as an argument, so portal is not needed.
    """
    plugin_obj = pas[plugin]
    activatable = []
    for info in plugin_obj.plugins.listPluginTypeInfo():
        interface = info['interface']
        interface_name = info['id']
        if plugin_obj.testImplements(interface) and \
                interface_name in selected_interfaces:
            if interface_name in disable:
                disable.append(interface_name)
                print >> out, " - Disabling: " + info['title']
            else:
                activatable.append(interface_name)
                print >> out, " - Activating: " + info['title']
    plugin_obj.manage_activateInterfaces(activatable)
    print >> out, plugin + " activated."

def movePluginToTop(pas, plugin_id, interface_name, out):
    """This moves a plugin to the top of the plugins list
    for a given interface.
    """
    registry = pas.plugins
    interface = registry._getInterfaceFromName(interface_name)
    while registry.listPlugins(interface)[0][0] != plugin_id:
        registry.movePluginsUp(interface,[plugin_id,])
    print >> out, "Moved " + plugin_id + " to top in " + interface_name + "."

def installSkin(portal, out, globals, skin_id, pos=-1):

    """
    Install a skin, and make sure it ends at the specified position in
    the portal_skins properties, so that it's easy to override
    existing skin elements with your product.
    """
    
    skins_tool = getToolByName(portal, 'portal_skins')

    if skin_id not in skins_tool.objectIds():
        addDirectoryViews(skins_tool, 'skins', globals)
        out.write("Added %s directory view to portal_skins\n" % skin_id)

    skins = skins_tool.getSkinSelections()
    for skin in skins:
        path = skins_tool.getSkinPath(skin)
        path = [p.strip() for p in path.split(',') if p]
        if skin_id in path:
            path.remove(skin_id)
        if path:
            path.insert(min(pos, len(path)), skin_id)
        else:
            path.insert(0, skin_id)
        skins_tool.addSkinSelection(skin, ','.join(path))
        
    print >> out, 'Installed skin'

def addTool(portal, product_name, tool_id):
    try:
        ctool = getToolByName(portal, tool_id)
    except AttributeError:
        portal.manage_addProduct[product_name].manage_addTool(product_name, 
                None)

def installConfiglets(portal, out, configlets, uninstall=False):

    """
    Install a configlet for the plone site. configlets parameter should be a 
    list of hashes.
    """

    print >> out, 'Installing configlet...'

    ctool = getToolByName(portal, 'portal_controlpanel')

    for c in configlets:
        print >> out, "-> installing configlet %s" % c['name']
        ctool.unregisterConfiglet(c['id'])
        if not uninstall:
            ctool.registerConfiglet(**c)

    print >> out, "done."
